@@ -1,34 +1,10 @@
 /*
- *
- *	RADIUS Accounting
- *	Remote Authentication Dial In User Service
- *
- *
- *	Livingston Enterprises, Inc.
- *	6920 Koll Center Parkway
- *	Pleasanton, CA   94566
- *
- *	Copyright 1992 - 1994 Livingston Enterprises, Inc.
- *	Copyright 1996 - 1997 Cistron Internet Services B.V.
- *
- *	Permission to use, copy, modify, and distribute this software for any
- *	purpose and without fee is hereby granted, provided that this
- *	copyright and permission notice appear on all copies and supporting
- *	documentation, the name of Livingston Enterprises, Inc. not be used
- *	in advertising or publicity pertaining to distribution of the
- *	program without specific prior permission, and notice be given
- *	in supporting documentation that copying and distribution is by
- *	permission of Livingston Enterprises, Inc.   
- *
- *	Livingston Enterprises, Inc. makes no representations about
- *	the suitability of this software for any purpose.  It is
- *	provided "as is" without express or implied warranty.
+ * acct.c	Accounting routines.
  *
  */
-
 char acct_sccsid[] =
-"@(#)acct.c	2.5  Copyright 1994 Livingston Enterprises Inc\n"
-"		     Copyright 1997 Cistron Internet Services B.V.";
+"@(#)acct.c	2.13 Copyright 1994 Livingston Enterprises Inc\n"
+"		     Copyright 1999 Cistron Internet Services B.V.";
 
 
 #include	<sys/types.h>
@@ -65,6 +41,33 @@ static char porttypes[] = "ASITX";
 
 #define LOCK_LEN sizeof(struct radutmp)
 
+/*
+ *	used for caching radutmp lookups.
+ */
+typedef struct nas_port {
+	UINT4			nasaddr;
+	int			port;
+	off_t			offset;
+	struct nas_port		*next;
+} NAS_PORT;
+static NAS_PORT *nas_port_list = NULL;
+
+
+/*
+ *	Lookup a NAS_PORT in the nas_port_list
+ */
+static NAS_PORT *nas_port_find(UINT4 nasaddr, int port)
+{
+	NAS_PORT	*cl;
+
+	for(cl = nas_port_list; cl; cl = cl->next)
+		if (nasaddr == cl->nasaddr &&
+			port == cl->port)
+			break;
+	return cl;
+}
+
+
 /*
  *	UUencode 4 bits base64. We use this to turn a 4 byte field
  *	(an IP adres) into 6 bytes of ASCII. This is used for the
@@ -76,13 +79,13 @@ char *uue(void *in)
 	static unsigned char res[7];
 	unsigned char *data = (char *)in;
 
-	res[0] = ENC(data[0] >> 2);
-	res[1] = ENC((data[0] << 4) & 060) + ((data[1] >> 4) & 017);
-	res[2] = ENC((data[1] << 2) & 074) + ((data[2] >> 6) & 03);
-	res[3] = ENC(data[2] & 077);
+	res[0] = ENC( data[0] >> 2 );
+	res[1] = ENC( ((data[0] << 4) & 060) + ((data[1] >> 4) & 017) );
+	res[2] = ENC( ((data[1] << 2) & 074) + ((data[2] >> 6) & 03) );
+	res[3] = ENC( data[2] & 077 );
 
-	res[4] = ENC(data[3] >> 2);
-	res[5] = ENC((data[3] << 4) & 060);
+	res[4] = ENC( data[3] >> 2 );
+	res[5] = ENC( (data[3] << 4) & 060 );
 	res[6] = 0;
 
 	for(i = 0; i < 6; i++) {
@@ -132,10 +135,12 @@ static void make_wtmp(struct radutmp *ut, struct utmp *wt, int status)
 	/*
 	 *	We store the dynamic IP address in the hostname field.
 	 */
+#ifdef UT_HOSTSIZE
 	if (ut->framed_address) {
 		ipaddr2str(buf, ntohl(ut->framed_address));
 		strncpy(wt->ut_host, buf, UT_HOSTSIZE);
 	}
+#endif
 #ifdef __svr4__
 	wt->ut_xtime = ut->time;
 #else
@@ -152,7 +157,7 @@ static void make_wtmp(struct radutmp *ut, struct utmp *wt, int status)
 		strcpy(wt->ut_id, "S");
 	else
 		strcpy(wt->ut_id, "T");
-	wt->ut_type = status == PW_STATUS_START ? USER_PROCESS : DEAD_PROCESS;
+	wt->ut_type = status == PW_STATUS_STOP ? DEAD_PROCESS : USER_PROCESS;
 #endif
 	if (status == PW_STATUS_STOP)
 		wt->ut_name[0] = 0;
@@ -198,9 +203,9 @@ int radzap(UINT4 nasaddr, int port, char *user, time_t t)
 			/*
 			 *	Match. Zap it.
 			 */
-			if (lseek(fd, -sizeof(u), SEEK_CUR) < 0) {
+			if (lseek(fd, -(off_t)sizeof(u), SEEK_CUR) < 0) {
 				log(L_ERR, "Accounting: radzap: negative lseek!\n");
-				lseek(fd, 0, SEEK_SET);
+				lseek(fd, (off_t)0, SEEK_SET);
 			}
 			u.type = P_IDLE;
 			u.time = t;
@@ -223,12 +228,9 @@ int radzap(UINT4 nasaddr, int port, char *user, time_t t)
 
 
 /*
- *	Show framed users in the RADIUS utmp file.
+ *	Store logins in the RADIUS utmp file.
  */
-int rad_accounting_new(authreq, activefd, dowtmp)
-AUTH_REQ	*authreq;
-int		activefd;
-int		dowtmp;
+int rad_accounting_new(AUTH_REQ *authreq, int dowtmp)
 {
 	struct radutmp	ut, u;
 	struct utmp	wt;
@@ -242,6 +244,10 @@ int		dowtmp;
 	time_t		t;
 	int		fd;
 	int		ret = 0;
+	int		just_an_update = 0;
+	int		port_seen = 0;
+	int		nas_port_type = 0;
+	int		off;
 
 	/*
 	 *	Which type is this.
@@ -255,7 +261,7 @@ int		dowtmp;
 	    status == PW_STATUS_ACCOUNTING_OFF) rb_record = 1;
 
 	if (!rb_record &&
-	    (vp = pairfind(authreq->request, PW_USER_NAME)) == NULL) {
+	    (vp = pairfind(authreq->request, PW_USER_NAME)) == NULL) do {
 		int check1 = 0;
 		int check2 = 0;
 
@@ -273,11 +279,16 @@ int		dowtmp;
 		     == NULL || vp->lvalue == 0)
 			check1 = 1;
 		if ((vp = pairfind(authreq->request, PW_ACCT_SESSION_ID))
-		     != NULL && memcmp(vp->strvalue, "00000000", 8) == 0)
+		     != NULL && vp->length == 8 &&
+		     memcmp(vp->strvalue, "00000000", 8) == 0)
 			check2 = 1;
 		if (check1 == 0 || check2 == 0) {
+#if 0 /* Cisco sometimes sends START records without username. */
 			log(L_ERR, "Accounting: no username in record");
 			return -1;
+#else
+			break;
+#endif
 		}
 		log(L_INFO, "Accounting: converting reboot records.");
 		if (status == PW_STATUS_STOP)
@@ -285,11 +296,11 @@ int		dowtmp;
 		if (status == PW_STATUS_START)
 			status = PW_STATUS_ACCOUNTING_ON;
 		rb_record = 1;
-	}
+	} while(0);
 
 #ifdef NT_DOMAIN_HACK
         if (!rb_record && vp) {
-		char buffer[128];
+		char buffer[AUTH_STRING_LEN];
 		char *ptr;
 		if ((ptr = strchr(vp->strvalue, '\\')) != NULL) {
 			strncpy(buffer, ptr + 1, sizeof(buffer));
@@ -302,7 +313,7 @@ int		dowtmp;
 	/*
 	 *	Add any specific attributes for this username.
 	 */
-	if (!rb_record) {
+	if (!rb_record && vp != NULL) {
 		hints_setup(authreq->request);
 		presuf_setup(authreq->request);
 	}
@@ -334,17 +345,30 @@ int		dowtmp;
 				break;
 			case PW_NAS_PORT_ID:
 				ut.nas_port = vp->lvalue;
+				port_seen = 1;
 				break;
 			case PW_ACCT_DELAY_TIME:
 				ut.delay = vp->lvalue;
 				break;
 			case PW_ACCT_SESSION_ID:
-				memcpy(ut.session_id, vp->strvalue,
+				/*
+				 *	If length > 8, only store the
+				 *	last 8 bytes.
+				 */
+				off = vp->length - sizeof(ut.session_id);
+				if (off < 0) off = 0;
+				memcpy(ut.session_id, vp->strvalue + off,
 					sizeof(ut.session_id));
 				break;
 			case PW_NAS_PORT_TYPE:
 				if (vp->lvalue >= 0 && vp->lvalue <= 4)
 					ut.porttype = porttypes[vp->lvalue];
+				nas_port_type = vp->lvalue;
+				break;
+			case PW_CALLING_STATION_ID:
+				strncpy(ut.caller_id, vp->strvalue,
+					sizeof(ut.caller_id));
+				ut.caller_id[sizeof(ut.caller_id)] = 0;
 				break;
 		}
 	}
@@ -386,16 +410,35 @@ int		dowtmp;
 	/*
 	 *	If we don't know this type of entry pretend we succeeded.
 	 */
-	if (status != PW_STATUS_START && status != PW_STATUS_STOP) {
+	if (status != PW_STATUS_START &&
+	    status != PW_STATUS_STOP &&
+	    status != PW_STATUS_ALIVE) {
 		log(L_ERR, "NAS %s port %d unknown packet type %d)",
 			nas_name(nas_address), ut.nas_port, status);
 		return 0;
 	}
 
+	/*
+	 *	Perhaps we don't want to store this record into
+	 *	radutmp/radwtmp. We skip records:
+	 *
+	 *	- without a NAS-Port-Id (telnet / tcp access)
+	 *	- with the username "!root" (console admin login)
+	 *	- with Port-Type = Sync (leased line up/down)
+	 */
+	if (!port_seen ||
+	    strncmp(ut.login, "!root", RUT_NAMESIZE) == 0
+#if 0 /* I HATE Ascend - they label ISDN as sync */
+	    || nas_port_type == PW_NAS_PORT_SYNC
+#endif
+	   )
+		return 0;
+
 	/*
 	 *	Enter into the radutmp file.
 	 */
 	if ((fd = open(RADUTMP, O_RDWR|O_CREAT, 0644)) >= 0) {
+		NAS_PORT *cache;
 		int r;
 
 		/*
@@ -409,11 +452,17 @@ int		dowtmp;
 	 	/*
 		 *	Find the entry for this NAS / portno combination.
 		 */
+		if ((cache = nas_port_find(ut.nas_address, ut.nas_port)) != NULL)
+			lseek(fd, (off_t)cache->offset, SEEK_SET);
+
 		r = 0;
+		off = 0;
 		while (read(fd, &u, sizeof(u)) == sizeof(u)) {
+			off += sizeof(u);
 			if (u.nas_address != ut.nas_address ||
 			    u.nas_port    != ut.nas_port)
 				continue;
+
 			if (status == PW_STATUS_STOP &&
 			    strncmp(ut.session_id, u.session_id, 
 			     sizeof(u.session_id)) != 0) {
@@ -429,25 +478,64 @@ int		dowtmp;
 				r = -1;
 				break;
 			}
+
 			if (status == PW_STATUS_START &&
 			    strncmp(ut.session_id, u.session_id, 
 			     sizeof(u.session_id)) == 0  &&
 			    u.time >= ut.time) {
+				if (u.type == P_LOGIN) {
+					log(L_INFO,
+		"Accounting: login: entry for NAS %s port %d duplicate",
+					nas_name(nas_address), u.nas_port);
+					r = -1;
+					dowtmp = 0;
+					break;
+				}
 				log(L_ERR,
-		"Accounting: login: entry for NAS %s port %d out of order",
+		"Accounting: login: entry for NAS %s port %d wrong order",
 				nas_name(nas_address), u.nas_port);
 				r = -1;
 				break;
 			}
-			if (lseek(fd, -sizeof(u), SEEK_CUR) < 0) {
-				log(L_ERR, "Accounting: negative lseek!\n");
-				lseek(fd, 0, SEEK_SET);
+
+			/*
+			 *	FIXME: the ALIVE record could need
+			 *	some more checking, but anyway I'd
+			 *	rather rewrite this mess -- miquels.
+			 */
+			if (status == PW_STATUS_ALIVE &&
+			    strncmp(ut.session_id, u.session_id, 
+			     sizeof(u.session_id)) == 0  &&
+			    u.type == P_LOGIN) {
+				/*
+				 *	Keep the original login time.
+				 */
+				ut.time = u.time;
+				if (u.login[0] != 0)
+					just_an_update = 1;
 			}
+
+			if (lseek(fd, -(off_t)sizeof(u), SEEK_CUR) < 0) {
+				log(L_ERR, "Accounting: negative lseek!\n");
+				lseek(fd, (off_t)0, SEEK_SET);
+				off = 0;
+			} else
+				off -= sizeof(u);
 			r = 1;
 			break;
 		}
 
-		if (r >= 0 && status == PW_STATUS_START) {
+		if (r >= 0 &&  (status == PW_STATUS_START ||
+				status == PW_STATUS_ALIVE)) {
+			if (cache == NULL) {
+			   if ((cache = malloc(sizeof(NAS_PORT))) != NULL) {
+				   cache->nasaddr = ut.nas_address;
+				   cache->port = ut.nas_port;
+				   cache->offset = off;
+				   cache->next = nas_port_list;
+				   nas_port_list = cache;
+			   }
+			}
 			ut.type = P_LOGIN;
 			write(fd, &ut, sizeof(u));
 		}
@@ -470,10 +558,18 @@ int		dowtmp;
 		ret = -1;
 	}
 
+	/*
+	 *	Don't write wtmp if we don't have a username, or
+	 *	if this is an update record and the original record
+	 *	was already written.
+	 */
+	if ((status != PW_STATUS_STOP && wt.ut_name[0] == 0) || just_an_update)
+		dowtmp = 0;
+
 	/*
 	 *	Write a RADIUS wtmp log file.
 	 */
-	if (doradwtmp && (fp = fopen(RADWTMP, "a")) != NULL) {
+	if (dowtmp && (fp = fopen(RADWTMP, "a")) != NULL) {
 		ret = 0;
 		fwrite(&wt, sizeof(wt), 1, fp);
 		fclose(fp);
@@ -485,7 +581,7 @@ int		dowtmp;
 
 
 
-int rad_accounting_orig(AUTH_REQ *authreq, int activefd, int authtype)
+int rad_accounting_orig(AUTH_REQ *authreq, int authtype, char *f)
 {
 	FILE		*outfd;
 	char		nasname[128];
@@ -514,18 +610,22 @@ int rad_accounting_orig(AUTH_REQ *authreq, int activefd, int authtype)
 	 */
 	cl = NULL;
 	nas = authreq->ipaddr;
-	if ((pair = pairfind(authreq->request, PW_NAS_IP_ADDRESS)) != NULL) {
+	if ((pair = pairfind(authreq->request, PW_NAS_IP_ADDRESS)) != NULL)
 		nas = pair->lvalue;
-		if ((cl = nas_find(nas)) != NULL) {
-			if (cl->shortname[0])
-				strcpy(nasname, cl->shortname);
-			else
-				strcpy(nasname, cl->longname);
-		}
+	if (authreq->server_ipaddr)
+		nas = authreq->server_ipaddr;
+
+	if ((cl = nas_find(nas)) != NULL) {
+		/* shortname and longname are < 128 bytes */
+		if (cl->shortname[0])
+			strcpy(nasname, cl->shortname);
+		else
+			strcpy(nasname, cl->longname);
 	}
+
 	if (cl == NULL) {
 		s = ip_hostname(nas);
-		if (strlen(s) > 127 || strchr(s, '/'))
+		if (strlen(s) >= sizeof(nasname) || strchr(s, '/'))
 			return -1;
 		strcpy(nasname, s);
 	}
@@ -533,17 +633,15 @@ int rad_accounting_orig(AUTH_REQ *authreq, int activefd, int authtype)
 	/*
 	 *	Create a directory for this nas.
 	 */
-	sprintf(buffer, "%s/%s", radacct_dir, nasname);
+	sprintf(buffer, "%.250s/%.250s", radacct_dir, nasname);
 	(void) mkdir(buffer, 0755);
 
 	/*
 	 *	Write Detail file.
 	 */
-	sprintf(buffer, "%s/%s/detail", radacct_dir, nasname);
+	sprintf(buffer, "%.200s/%.200s/%.100s", radacct_dir, nasname, f ? f : "detail");
 	if((outfd = fopen(buffer, "a")) == (FILE *)NULL) {
-		log(L_ERR,
-			"Acct: Couldn't open file %s/%s/detail",
-				radacct_dir, nasname);
+		log(L_ERR, "Acct: Couldn't open file %s", buffer);
 		ret = -1;
 	} else {
 
@@ -553,9 +651,11 @@ int rad_accounting_orig(AUTH_REQ *authreq, int activefd, int authtype)
 		/* Write each attribute/value to the log file */
 		pair = authreq->request;
 		while(pair != (VALUE_PAIR *)NULL) {
-			fputs("\t", outfd);
-			fprint_attr_val(outfd, pair);
-			fputs("\n", outfd);
+			if (pair->attribute != PW_PASSWORD) {
+				fputs("\t", outfd);
+				fprint_attr_val(outfd, pair);
+				fputs("\n", outfd);
+			}
 			pair = pair->next;
 		}
 
@@ -583,157 +683,11 @@ int rad_accounting_orig(AUTH_REQ *authreq, int activefd, int authtype)
 	return ret;
 }
 
-/*
- *	Reply to the request with an ACKNOWLEDGE.  Also attach
- *	reply attribute value pairs and any user message provided.
- */
-static int send_acct_reply(AUTH_REQ *authreq, VALUE_PAIR *reply,
-		char *msg, int activefd)
-{
-	AUTH_HDR		*auth;
-	u_short			total_length;
-	struct	sockaddr_in	saremote;
-	struct	sockaddr_in	*sin;
-	u_char			*ptr;
-	int			len;
-	UINT4			lvalue;
-	u_char			digest[16];
-	int			secretlen;
-
-	auth = (AUTH_HDR *)send_buffer;
-
-	/* Build standard header */
-	auth->code = PW_ACCOUNTING_RESPONSE;
-	auth->id = authreq->id;
-	memcpy(auth->vector, authreq->vector, AUTH_VECTOR_LEN);
-
-	DEBUG("Sending Accounting Ack of id %d to %lx (%s)\n",
-	    authreq->id, (u_long)authreq->ipaddr, ip_hostname(authreq->ipaddr));
-
-	total_length = AUTH_HDR_LEN;
-
-	/* Load up the configuration values for the user */
-	ptr = auth->data;
-	while(reply != (VALUE_PAIR *)NULL) {
-		debug_pair(stdout, reply);
-		*ptr++ = reply->attribute;
-
-		switch(reply->type) {
-
-		case PW_TYPE_STRING:
-			len = strlen(reply->strvalue);
-			*ptr++ = len + 2;
-			strcpy(ptr, reply->strvalue);
-			ptr += len;
-			total_length += len + 2;
-			break;
-			
-		case PW_TYPE_INTEGER:
-		case PW_TYPE_IPADDR:
-			*ptr++ = sizeof(UINT4) + 2;
-			lvalue = htonl(reply->lvalue);
-			memcpy(ptr, &lvalue, sizeof(UINT4));
-			ptr += sizeof(UINT4);
-			total_length += sizeof(UINT4) + 2;
-			break;
-
-		default:
-			break;
-		}
-
-		reply = reply->next;
-	}
-
-	/* Append the user message */
-	if(msg != (char *)NULL) {
-		len = strlen(msg);
-		if(len > 0 && len < AUTH_STRING_LEN) {
-			*ptr++ = PW_REPLY_MESSAGE;
-			*ptr++ = len + 2;
-			memcpy(ptr, msg, len);
-			ptr += len;
-			total_length += len + 2;
-		}
-	}
-
-	auth->length = htons(total_length);
-
-	/* Calculate the response digest */
-	secretlen = strlen(authreq->secret);
-	memcpy(send_buffer + total_length, authreq->secret, secretlen);
-	md5_calc(digest, (char *)auth, total_length + secretlen);
-	memcpy(auth->vector, digest, AUTH_VECTOR_LEN);
-	memset(send_buffer + total_length, 0, secretlen);
-
-	sin = (struct sockaddr_in *) &saremote;
-        memset ((char *) sin, '\0', sizeof (saremote));
-	sin->sin_family = AF_INET;
-	sin->sin_addr.s_addr = htonl(authreq->ipaddr);
-	sin->sin_port = htons(authreq->udp_port);
-
-	/* Send it to the user */
-	sendto(activefd, (char *)auth, (int)total_length, (int)0,
-			(struct sockaddr *) sin, sizeof(struct sockaddr_in));
-
-	return 0;
-}
-
-/*
- *	Validates the requesting client NAS.  Calculates the
- *	signature based on the clients private key.
- */
-static int calc_acctdigest(u_char *digest, AUTH_REQ *authreq,
-			char *recvbuf, int len)
-{
-	int	secretlen;
-	CLIENT	*cl;
-	char zero[AUTH_VECTOR_LEN];
-
-	/*
-	 *	See if we know this client.
-	 */
-	if ((cl = client_find(authreq->ipaddr)) == NULL) {
-		log(L_ERR, "request from unknown client: %s",
-			client_name(authreq->ipaddr));
-		return -1;
-	}
-
-	/*
-	 *	Copy secret into authreq->secret so that we can
-	 *	use it with send_acct_reply()
-	 */
-	secretlen = strlen(cl->secret);
-	strcpy(authreq->secret, cl->secret);
-
-	/*
-	 *	Older clients have the authentication vector set to
-	 *	all zeros. Return `1' in that case.
-	 */
-	memset(zero, 0, sizeof(zero));
-	if (memcmp(authreq->vector, zero, AUTH_VECTOR_LEN) == 0)
-		return 1;
-
-	/*
-	 *	Zero out the auth_vector in the received packet.
-	 *	Then append the shared secret to the received packet,
-	 *	and calculate the MD5 sum. This must be the same
-	 *	as the original MD5 sum (authreq->vector).
-	 */
-	memset(recvbuf + 4, 0, AUTH_VECTOR_LEN);
-	memcpy(recvbuf + len, cl->secret, secretlen);
-	md5_calc(digest, recvbuf, len + secretlen);
-
-	/*
-	 *	Return 0 if OK, 2 if not OK.
-	 */
-	return memcmp(digest, authreq->vector, AUTH_VECTOR_LEN) ? 2 : 0;
-}
-
 
 /*
  *	rad_accounting: call both the old and new style accounting functions.
  */
-int rad_accounting(AUTH_REQ *authreq, int activefd, char *recvbuf, int len)
+int rad_accounting(AUTH_REQ *authreq, int activefd)
 {
 	int reply = 0;
 	int auth;
@@ -743,11 +697,10 @@ int rad_accounting(AUTH_REQ *authreq, int activefd, char *recvbuf, int len)
 	 *	See if we know this client, then check the
 	 *	request authenticator.
 	 */
-	auth = calc_acctdigest(pw_digest, authreq, recvbuf, len);
+	auth = calc_acctdigest(pw_digest, authreq);
+
 	if (auth < 0) {
-		pairfree(authreq->request);
-		memset(authreq, 0, sizeof(AUTH_REQ));
-		free(authreq);
+		authfree(authreq);
 		return -1;
 	}
 
@@ -757,18 +710,18 @@ int rad_accounting(AUTH_REQ *authreq, int activefd, char *recvbuf, int len)
 		 *	we run it first, the stripped info will also
 		 *	get into the "details" file.
 		 */
-		if (rad_accounting_new(authreq, activefd, 1) == 0)
+		if (rad_accounting_new(authreq, doradwtmp) == 0)
 			reply = 1;
-		if (rad_accounting_orig(authreq, activefd, auth) == 0)
+		if (rad_accounting_orig(authreq, auth, NULL) == 0)
 			reply = 1;
 	} else {
 		/*
 		 *	First log into the details file, before the
 		 *	username gets stripped by rad_accounting_new.
 		 */
-		if (rad_accounting_orig(authreq, activefd, auth) == 0)
+		if (rad_accounting_orig(authreq, auth, NULL) == 0)
 			reply = 1;
-		if (rad_accounting_new(authreq, activefd, 1) == 0)
+		if (rad_accounting_new(authreq, doradwtmp) == 0)
 			reply = 1;
 	}
 
@@ -776,64 +729,16 @@ int rad_accounting(AUTH_REQ *authreq, int activefd, char *recvbuf, int len)
 		/*
 		 *	Now send back an ACK to the NAS.
 		 */
-		send_acct_reply(authreq, (VALUE_PAIR *)NULL, 
-				(char *)NULL,activefd);
+		rad_send_reply(PW_ACCOUNTING_RESPONSE,
+			authreq, NULL, NULL, activefd);
 	}
 
-	pairfree(authreq->request);
-	memset(authreq, 0, sizeof(AUTH_REQ));
-	free(authreq);
+	authfree(authreq);
 
 	return reply ? 0 : -1;
 }
 
 
-#if 0 /* UNRELIABLE */
-/*
- *	See if a user is already logged in.
- *
- *	For now we implement another algorithm. If the total no. of logins
- *	exceeds the maximum, we re-read the utmp file. We keep on trying
- *	for 30 seconds. This does not solve the "lost accounting records"
- *	problem, but it does fix the problem with delayed logout records.
- */
-int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul)
-{
-	FILE		*fp;
-	int		count;
-	struct radutmp	u;
-	int		tries = 0;
-
-	if ((fp = fopen(RADUTMP, "r")) < 0)
-		return 0;
-
-	while(tries < 15) {
-
-		if (tries++ > 0) sleep(2);
-
-		/*
-		 *	FIXME: lockf() the file while reading?
-		 */
-		count = 0;
-		rewind(fp);
-		while(fread(&u, sizeof(u), 1, fp) == 1)
-			if (strncmp(name, u.login, RUT_NAMESIZE) == 0
-			    && u.type == P_LOGIN)
-				count++;
-
-
-		if (count < maxsimul) {
-			fclose(fp);
-			return 0;
-		}
-	}
-	fclose(fp);
-
-	return 1;
-}
-
-#else /* RELIABLE */
-
 /*
  *	Timeout handler (10 secs)
  */
@@ -855,7 +760,7 @@ static int rad_check_ts(struct radutmp *ut)
 	char	port[8];
 	char	session_id[12];
 	char	*s;
-	void    (*handler)(int);
+	void	(*handler)(int);
 
 	/*
 	 *	Find NAS type.
@@ -877,13 +782,13 @@ static int rad_check_ts(struct radutmp *ut)
 
 	if (pid > 0) {
 		/*
-		 *	Parent - Wait for checklogin to terminate.
+		 *	Parent - Wait for checkrad to terminate.
 		 *	We timeout in 10 seconds.
 		 */
 		got_alrm = 0;
 		signal(SIGALRM, alrm_handler);
 		alarm(10);
-		while((e = wait(&st)) != pid)
+		while((e = waitpid(pid, &st, 0)) != pid)
 			if (e < 0 && (errno != EINTR || got_alrm))
 				break;
 		alarm(0);
@@ -942,7 +847,7 @@ int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul)
 	FILE		*wfp;
 	int		fd;
 	int		count;
-	struct radutmp	u, empty;
+	struct radutmp	u;
 	struct utmp	wt;
 	VALUE_PAIR	*fra;
 	int		mpp = 1;
@@ -964,14 +869,13 @@ int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul)
 		close(fd);
 		return 0;
 	}
-	lseek(fd, 0L, SEEK_SET);
+	lseek(fd, (off_t)0, SEEK_SET);
 
 	/*
 	 *	Setup some stuff, like for MPP detection.
 	 */
 	if ((fra = pairfind(request, PW_FRAMED_IP_ADDRESS)) != NULL)
 		ipno = htonl(fra->lvalue);
-	memset(&empty, 0, sizeof(empty));
 
 	/*
 	 *	lockf() the file while reading/writing.
@@ -1005,8 +909,9 @@ int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul)
 				 *	False record - zap it.
 				 */
 
-				lseek(fd, -sizeof(u), SEEK_CUR);
-				write(fd, &empty, sizeof(empty));
+				lseek(fd, -(off_t)sizeof(u), SEEK_CUR);
+				u.type = P_IDLE;
+				write(fd, &u, sizeof(u));
 
 				if ((wfp = fopen(RADWTMP, "a")) != NULL) {
 					make_wtmp(&u, &wt, PW_STATUS_STOP);
@@ -1020,5 +925,4 @@ int rad_check_multi(char *name, VALUE_PAIR *request, int maxsimul)
 
 	return (count < maxsimul) ? 0 : mpp;
 }
-#endif
 
