@@ -9,7 +9,7 @@
  *	Pleasanton, CA   94566
  *
  *	Copyright 1992 Livingston Enterprises, Inc.
- *	Copyright 1996-1997 Cistron Internet Services B.V.
+ *	Copyright 1996-1999 Cistron Internet Services B.V.
  *
  *	Permission to use, copy, modify, and distribute this software for any
  *	purpose and without fee is hereby granted, provided that this
@@ -28,11 +28,12 @@
 
 char files_sccsid[] =
 "@(#)users.c	1.12 Copyright 1992 Livingston Enterprises Inc\n"
-"@(#)files.c	2.27 Copyright 1997 Cistron Internet Services B.V.";
+"@(#)files.c	2.42 Copyright 1999 Cistron Internet Services B.V.";
 
 #include	<sys/types.h>
 #include	<sys/socket.h>
 #include	<sys/time.h>
+#include	<sys/stat.h>
 #include	<netinet/in.h>
 
 #include	<stdio.h>
@@ -53,16 +54,18 @@ char files_sccsid[] =
 
 #include	"radiusd.h"
 
-static	int userparse(char *buffer, VALUE_PAIR **first_pair);
 static	int  huntgroup_match(VALUE_PAIR *, char *);
 #ifdef NDBM
 static	DBM *dbmfile;
 #endif
 
+#include "cache.h"
+
 struct pair_list {
 	char *name;
 	VALUE_PAIR *check;
 	VALUE_PAIR *reply;
+	int lineno;
 	struct pair_list *next;
 };
 typedef struct pair_list PAIR_LIST;
@@ -76,6 +79,8 @@ PAIR_LIST	*huntgroups;
 PAIR_LIST	*hints;
 CLIENT		*clients;
 NAS		*naslist;
+REALM		*realms;
+
 
 /*
  *	Returns a copy of a pair list.
@@ -106,10 +111,10 @@ static VALUE_PAIR *paircopy(VALUE_PAIR *from)
  *	Move attributes from one list to the other
  *	if not already present.
  */
-static void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
+void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
 {
-	VALUE_PAIR *tail, *i, *next;
-	VALUE_PAIR *last = NULL;
+	VALUE_PAIR *tailto, *i, *next;
+	VALUE_PAIR *tailfrom = NULL;
 	int has_password = 0;
 
 	if (*to == NULL) {
@@ -118,37 +123,64 @@ static void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
 		return;
 	}
 
-	tail = *to;
+	/*
+	 *	First, see if there are any passwords here, and
+	 *	point "tailto" to the end of the "to" list.
+	 */
+	tailto = *to;
 	for(i = *to; i; i = i->next) {
 		if (i->attribute == PW_PASSWORD ||
-		    i->attribute == PW_CHAP_PASSWORD ||
+		/*
+		 *	FIXME: this seems to be needed with PAM support
+		 *	to keep it around the Auth-Type = Pam stuff.
+		 *	Perhaps we should only do this if Auth-Type = Pam?
+		 */
+#ifdef PAM
+		    i->attribute == PAM_AUTH_ATTR ||
+#endif
 		    i->attribute == PW_CRYPT_PASSWORD)
 			has_password = 1;
-		tail = i;
+		tailto = i;
 	}
 
+	/*
+	 *	Loop over the "from" list.
+	 */
 	for(i = *from; i; i = next) {
 		next = i->next;
+		/*
+		 *	If there was a password in the "to" list,
+		 *	do not move any other password from the
+		 *	"from" to the "to" list.
+		 */
 		if (has_password &&
 		    (i->attribute == PW_PASSWORD ||
-		     i->attribute == PW_CHAP_PASSWORD ||
+#ifdef PAM
+		     i->attribute == PAM_AUTH_ATTR ||
+#endif
 		     i->attribute == PW_CRYPT_PASSWORD)) {
-			last = i;
+			tailfrom = i;
 			continue;
 		}
+		/*
+		 *	If the attribute is already present in "to",
+		 *	do not move it from "from" to "to". We make
+		 *	an exception for "Hint" which can appear multiple
+		 *	times, and we never move "Fall-Through".
+		 */
 		if (i->attribute == PW_FALL_THROUGH ||
-		    (i->attribute != PW_HINT &&
-		     pairfind(*to, i->attribute) != 0)) {
-			last = i;
+		    (i->attribute != PW_HINT && i->attribute != PW_FRAMED_ROUTE
+		     && pairfind(*to, i->attribute) != 0)) {
+			tailfrom = i;
 			continue;
 		}
-		if (last)
-			last->next = next;
+		if (tailfrom)
+			tailfrom->next = next;
 		else
 			*from = next;
-		tail->next = i;
+		tailto->next = i;
 		i->next = NULL;
-		tail = i;
+		tailto = i;
 	}
 }
 
@@ -157,29 +189,44 @@ static void pairmove(VALUE_PAIR **to, VALUE_PAIR **from)
  */
 void pairmove2(VALUE_PAIR **to, VALUE_PAIR **from, int attr)
 {
-	VALUE_PAIR *tail, *i, *next;
-	VALUE_PAIR *last = NULL;
+	VALUE_PAIR *to_tail, *i, *next;
+	VALUE_PAIR *iprev = NULL;
 
-	if (*to == NULL)
-		return;
-
-	tail = *to;
-	for(i = *to; i; i = i->next)
-		tail = i;
+	/*
+	 *	Find the last pair in the "to" list and put it in "to_tail".
+	 */
+	if (*to != NULL) {
+		to_tail = *to;
+		for(i = *to; i; i = i->next)
+			to_tail = i;
+	} else
+		to_tail = NULL;
 
 	for(i = *from; i; i = next) {
 		next = i->next;
 
-		if (i->attribute != attr)
+		if (i->attribute != attr) {
+			iprev = i;
 			continue;
+		}
 
-		if (last)
-			last->next = next;
+		/*
+		 *	Remove the attribute from the "from" list.
+		 */
+		if (iprev)
+			iprev->next = next;
 		else
 			*from = next;
-		tail->next = i;
+
+		/*
+		 *	Add the attribute to the "to" list.
+		 */
+		if (to_tail)
+			to_tail->next = i;
+		else
+			*to = i;
+		to_tail = i;
 		i->next = NULL;
-		tail = i;
 	}
 }
 
@@ -236,7 +283,7 @@ void presuf_setup(VALUE_PAIR *request_pairs)
 		if ((presuf_pair = pairfind(pl->check, PW_PREFIX)) == NULL &&
 		    (presuf_pair = pairfind(pl->check, PW_SUFFIX)) == NULL)
 			continue;
-		if (presufcmp(presuf_pair, name_pair->strvalue, name) != 0)
+		if (presufcmp(presuf_pair, name_pair->strvalue, name, sizeof(name)) != 0)
 			continue;
 		/*
 		 *	See if username must be stripped.
@@ -244,7 +291,9 @@ void presuf_setup(VALUE_PAIR *request_pairs)
 		if ((tmp = pairfind(pl->check, PW_STRIP_USERNAME)) != NULL &&
 		    tmp->lvalue == 0)
 			continue;
+		/* sizeof(name_pair->strvalue) > sizeof(name) */
 		strcpy(name_pair->strvalue, name);
+		name_pair->length = strlen(name_pair->strvalue);
 		break;
 	}
 }
@@ -256,11 +305,12 @@ void presuf_setup(VALUE_PAIR *request_pairs)
  */
 static int portcmp(VALUE_PAIR *check, VALUE_PAIR *request)
 {
-	char buf[256];
+	char buf[AUTH_STRING_LEN];
 	char *s, *p;
 	int lo, hi;
 	int port = request->lvalue;
 
+	/* Same size */
 	strcpy(buf, check->strvalue);
 	s = strtok(buf, ",");
 	while(s) {
@@ -282,27 +332,40 @@ static int portcmp(VALUE_PAIR *check, VALUE_PAIR *request)
 
 /*
  *	See if user is member of a group.
- *	FIXME:  We don't handle additional groups yet.
+ *	We also handle additional groups.
  */
-static int groupcmp(VALUE_PAIR *check, VALUE_PAIR *request)
+static int groupcmp(VALUE_PAIR *check, char *username)
 {
 	struct passwd *pwd;
 	struct group *grp;
+	char **member;
+	int retval;
 
-	if ((pwd = rad_getpwnam(request->strvalue)) == NULL)
+	if (cache_passwd && (retval = H_groupcmp(check, username)) != -2)
+		return retval;
+
+	if ((pwd = rad_getpwnam(username)) == NULL)
 		return -1;
 
 	if ((grp = getgrnam(check->strvalue)) == NULL)
 		return -1;
 
-	return (pwd->pw_gid == grp->gr_gid) ? 0 : -1;
+	retval = (pwd->pw_gid == grp->gr_gid) ? 0 : -1;
+	if (retval < 0) {
+		for (member = grp->gr_mem; *member && retval; member++) {
+			if (strcmp(*member, pwd->pw_name) == 0)
+				retval = 0;
+		}
+	}
+	return retval;
 }
 
 /*
  *	Compare prefix/suffix.
  */
-int presufcmp(VALUE_PAIR *check, char *name, char *rest)
+int presufcmp(VALUE_PAIR *check, char *name, char *rest, int restlen)
 {
+	char *realm;
 	int len, namelen;
 	int ret = -1;
 
@@ -311,12 +374,22 @@ int presufcmp(VALUE_PAIR *check, char *name, char *rest)
 		name, check->strvalue, check->attribute);
 #endif
 
+	/*
+	 *	Temporarily strip off the realm, unless ofcourse
+	 *	we explicitly want to compare against a realm.
+	 */
+	realm = NULL;
+	if (strchr(check->strvalue, '@') == NULL) {
+		realm = strchr(name, '@');
+		if (realm) *realm = 0;
+	}
+
 	len = strlen(check->strvalue);
 	switch (check->attribute) {
 		case PW_PREFIX:
 			ret = strncmp(name, check->strvalue, len);
 			if (ret == 0 && rest)
-				strcpy(rest, name + len);
+				strNcpy(rest, name + len, restlen);
 			break;
 		case PW_SUFFIX:
 			namelen = strlen(name);
@@ -329,6 +402,13 @@ int presufcmp(VALUE_PAIR *check, char *name, char *rest)
 			}
 			break;
 	}
+
+	if (realm) *realm = '@';
+	if (ret == 0 && rest && realm) {
+		len = strlen(rest);
+		strNcpy(rest + len, realm, restlen -len);
+	}
+
 	return ret;
 }
 
@@ -343,7 +423,10 @@ static int paircmp(VALUE_PAIR *request, VALUE_PAIR *check)
 	VALUE_PAIR *check_item = check;
 	VALUE_PAIR *auth_item;
 	int result = 0;
+	char username[AUTH_STRING_LEN];
+	int compare;
 
+	username[0] = '\0';
 	while (result == 0 && check_item != NULL) {
 		switch (check_item->attribute) {
 			/*
@@ -351,10 +434,13 @@ static int paircmp(VALUE_PAIR *request, VALUE_PAIR *check)
 			 *	These are "server" check items.
 			 */
 			case PW_EXPIRATION:
+			case PW_LOGIN_TIME:
 			case PW_PASSWORD:
-			case PW_CHAP_PASSWORD:
 			case PW_CRYPT_PASSWORD:
 			case PW_AUTHTYPE:
+#ifdef PAM /* cjd 19980706 */
+                        case PAM_AUTH_ATTR:
+#endif
 			case PW_SIMULTANEOUS_USE:
 			case PW_STRIP_USERNAME:
 				check_item = check_item->next;
@@ -373,6 +459,8 @@ static int paircmp(VALUE_PAIR *request, VALUE_PAIR *check)
 					if (auth_item->attribute  !=
 					    PW_USER_NAME)
 						continue;
+					/* Sizes are the same */
+					strcpy(username, auth_item->strvalue);
 				case PW_HUNTGROUP_NAME:
 					break;
 				case PW_HINT:
@@ -399,50 +487,78 @@ static int paircmp(VALUE_PAIR *request, VALUE_PAIR *check)
 		 *	OK it is present now compare them.
 		 */
 		
+		compare = 0;	/* default result */
 		switch(check_item->type) {
-
 			case PW_TYPE_STRING:
 				if (check_item->attribute == PW_PREFIX ||
 				    check_item->attribute == PW_SUFFIX) {
 					if (presufcmp(check_item,
-					    auth_item->strvalue, NULL) != 0)
-						result = -1;
+					    auth_item->strvalue, username,
+					    sizeof(username)) != 0)
+						return -1;
 				}
 				else
 				if (check_item->attribute == PW_NAS_PORT_ID) {
 					if (portcmp(check_item, auth_item) != 0)
-						result = -1;
+						return -1;
 				}
 			 	else
 				if (check_item->attribute == PW_GROUP_NAME ||
 				    check_item->attribute == PW_GROUP) {
-					if (groupcmp(check_item, auth_item)!=0)
-						result = -1;
+					compare = groupcmp(check_item, username);
 				}
 				else
 				if (check_item->attribute == PW_HUNTGROUP_NAME){
-					if (!huntgroup_match(request,
-						check_item->strvalue))
-							result = -1;
+					compare = (huntgroup_match(request,
+						check_item->strvalue) == 0);
+					DEBUG("COMPARE = %d", compare);
+					break;
 				}
 				else
-				if (strcmp(check_item->strvalue,
-						auth_item->strvalue) != 0) {
-					result = -1;
-				}
+				compare = strcmp(auth_item->strvalue,
+						 check_item->strvalue);
 				break;
 
 			case PW_TYPE_INTEGER:
 			case PW_TYPE_IPADDR:
-				if(check_item->lvalue != auth_item->lvalue) {
-					result = -1;
-				}
+				compare = auth_item->lvalue - check_item->lvalue;
 				break;
 
 			default:
-				result = -1;
+				return -1;
 				break;
 		}
+
+		switch (check_item->operator)
+		  {
+		  default:
+		  case PW_OPERATOR_EQUAL:
+		    if (compare != 0) return -1;
+		    break;
+
+		  case PW_OPERATOR_NOT_EQUAL:
+		    if (compare == 0) return -1;
+		    break;
+
+		  case PW_OPERATOR_LESS_THAN:
+		    if (compare >= 0) return -1;
+		    break;
+
+		  case PW_OPERATOR_GREATER_THAN:
+		    if (compare <= 0) return -1;
+		    break;
+		    
+		  case PW_OPERATOR_LESS_EQUAL:
+		    if (compare > 0) return -1;
+		    break;
+
+		  case PW_OPERATOR_GREATER_EQUAL:
+		    if (compare < 0) return -1;
+		    break;
+		  }
+		
+
+
 		if (result == 0)
 			check_item = check_item->next;
 	}
@@ -472,6 +588,11 @@ static int hunt_paircmp(VALUE_PAIR *request, VALUE_PAIR *check)
 			case PW_EXPIRATION:
 			case PW_PASSWORD:
 			case PW_AUTHTYPE:
+/* cjd 19980706
+ */
+#ifdef PAM
+                        case PAM_AUTH_ATTR:
+#endif
 			case PW_SIMULTANEOUS_USE:
 				check_item = check_item->next;
 				continue;
@@ -507,8 +628,8 @@ static int hunt_paircmp(VALUE_PAIR *request, VALUE_PAIR *check)
 			 	else
 				if (check_item->attribute == PW_GROUP_NAME ||
 				    check_item->attribute == PW_GROUP) {
-					if (groupcmp(check_item, auth_item)==0)
-						result = 0;
+					if (groupcmp(check_item, auth_item->strvalue)==0)
+							result = 0;
 				}
 				else
 				if (check_item->attribute == PW_HUNTGROUP_NAME){
@@ -586,6 +707,99 @@ static char *getusername(char *buffer, char *entry)
 	return ptr;
 }
 
+
+/*
+ *	Free a PAIR_LIST
+ */
+static void pairlist_free(PAIR_LIST **pl)
+{
+	PAIR_LIST *p, *next;
+
+	for (p = *pl; p; p = next) {
+		if (p->name) free(p->name);
+		if (p->check) pairfree(p->check);
+		if (p->reply) pairfree(p->reply);
+		next = p->next;
+		free(p);
+	}
+	*pl = NULL;
+}
+
+/*
+ *	Fixup a check line.
+ *	If Password or Crypt-Password is set, but there is no
+ *	Auth-Type, add one (kludge!).
+ */
+static void auth_type_fixup(VALUE_PAIR *check)
+{
+	VALUE_PAIR	*vp;
+	VALUE_PAIR	*c = NULL;
+	int		n;
+
+	/*
+	 *	See if a password is present. Return right away
+	 *	if we see Auth-Type.
+	 */
+	for (vp = check; vp; vp = vp->next) {
+		if (vp->attribute == PW_AUTHTYPE)
+			return;
+		if (vp->attribute == PW_PASSWORD) {
+			c = vp;
+			n = PW_AUTHTYPE_LOCAL;
+		}
+		if (vp->attribute == PW_CRYPT_PASSWORD) {
+			c = vp;
+			n = PW_AUTHTYPE_CRYPT;
+		}
+	}
+
+	if (c == NULL)
+		return;
+
+	/*
+	 *	Add an Auth-Type attribute.
+	 *	FIXME: put Auth-Type _first_ (doesn't matter now,
+	 *	might matter some day).
+	 *	
+	 */
+	if ((vp = malloc(sizeof(VALUE_PAIR))) == NULL) {
+		log(L_CONS|L_ERR, "no memory");
+		exit(1);
+	}
+	memset(vp, 0, sizeof(VALUE_PAIR));
+	strcpy(vp->name, "Auth-Type");
+	vp->attribute = PW_AUTHTYPE;
+	vp->type = PW_TYPE_INTEGER;
+	vp->lvalue = n;
+
+	vp->next = c->next;
+	c->next = vp;
+
+}
+
+static void pair_sanity(char *fn, PAIR_LIST *entry)
+{
+	VALUE_PAIR *pair;
+
+	for (entry = users; entry; entry = entry->next) {
+		for (pair = entry->reply; pair; pair = pair->next) {
+			if ((pair->attribute == PW_PASSWORD) ||
+			    (pair->attribute == PW_CHAP_PASSWORD) ||
+			    (pair->attribute >= 1000 &&
+			     pair->attribute <= 65536)) {
+
+		log(L_ERR|L_CONS, "\n[%s:%d] WARNING: Check item \"%s\"\n"
+		    "\tfound in reply item list for user \"%s\".\n"
+		    "\tThis attribute MUST go on the first line"
+		    " with the other check items.\n", fn, entry->lineno,
+		    pair->name, entry->name);
+
+			}
+		}
+	}
+}
+
+
 #define FIND_MODE_NAME	0
 #define FIND_MODE_REPLY	1
 
@@ -604,6 +818,7 @@ static PAIR_LIST *file_read(char *file, int complain)
 	VALUE_PAIR	*reply_tmp;
 	PAIR_LIST	*pl = NULL, *last = NULL, *t;
 	int		lineno = 0;
+	int		old_lineno;
 
 	/*
 	 *	Open the table
@@ -668,6 +883,7 @@ parse_again:
 			 */
 			check_tmp = NULL;
 			reply_tmp = NULL;
+			old_lineno = lineno;
 			if(userparse(ptr, &check_tmp) != 0) {
 				log(L_ERR|L_CONS,
 				"%s[%d]: Parse error (check) for entry %s",
@@ -698,10 +914,12 @@ parse_again:
 					perror(progname);
 					exit(1);
 				}
+				auth_type_fixup(check_tmp);
 				memset(t, 0, sizeof(*t));
 				t->name = strdup(entry);
 				t->check = check_tmp;
 				t->reply = reply_tmp;
+				t->lineno = old_lineno;
 				check_tmp = NULL;
 				reply_tmp = NULL;
 				if (last)
@@ -710,6 +928,8 @@ parse_again:
 					pl = t;
 				last = t;
 
+				pair_sanity(file, t);
+
 				mode = FIND_MODE_NAME;
 				if (buffer[0] != 0)
 					goto parse_again;
@@ -795,9 +1015,11 @@ static int dbm_find(char *name, VALUE_PAIR *request_pairs,
 	/*
 	 *	See if the check_pairs match.
 	 */
+	auth_type_fixup(check_tmp);
 	if (paircmp(request_pairs, check_tmp) == 0) {
 		ret = 1;
 		pairmove(reply_pairs, &reply_tmp);
+		pairmove2(reply_pairs, &reply_tmp, PW_FALL_THROUGH);
 		pairmove(check_pairs, &check_tmp);
 	}
 	pairfree(reply_tmp);
@@ -819,6 +1041,31 @@ static int fallthrough(VALUE_PAIR *vp)
 	return tmp ? tmp->lvalue : 0;
 }
 
+
+#ifdef ASCEND_PORT_HACK
+/*
+ *	dgreer --
+ *	This hack changes Ascend's wierd port numberings
+ *      to standard 0-??? port numbers so that the "+" works
+ *      for IP address assignments.
+ */
+static int ascend_port_number(int nas_port)
+{
+	int service;
+	int line;
+	int channel;
+
+	if (nas_port > 9999) {
+		service = nas_port/10000; /* 1=digital 2=analog */
+		line = (nas_port - (10000 * service)) / 100;
+		channel = nas_port-((10000 * service)+(100 * line));
+		nas_port =
+			(channel - 1) + (line - 1) * ASCEND_CHANNELS_PER_LINE;
+	}
+	return nas_port;
+}
+#endif
+
 /*
  *	Find the named user in the database.  Create the
  *	set of attribute-value pairs to check and reply with
@@ -833,14 +1080,12 @@ int user_find(char *name, VALUE_PAIR *request_pairs,
 	VALUE_PAIR	*reply_tmp;
 	VALUE_PAIR	*tmp;
 	PAIR_LIST	*pl;
+	int		found = 0;
 #ifdef USE_DBM
 	int		i, r;
 	char		buffer[256];
 #endif
 
-	*check_pairs = NULL;
-	*reply_pairs = NULL;
-
 	/* 
 	 *	Check for valid input, zero length names not permitted 
 	 */
@@ -854,7 +1099,7 @@ int user_find(char *name, VALUE_PAIR *request_pairs,
 	 */
 	if ((tmp = pairfind(request_pairs, PW_NAS_PORT_ID)) != NULL)
 		nas_port = tmp->lvalue;
-		
+
 	/*
 	 *	Find the entry for the user.
 	 */
@@ -863,7 +1108,7 @@ int user_find(char *name, VALUE_PAIR *request_pairs,
 		/*
 		 *	FIXME: No Prefix / Suffix support for DBM.
 		 */
-		sprintf(buffer, "%s/%s", radius_dir, RADIUS_USERS);
+		sprintf(buffer, "%.200s/%.50s", radius_dir, RADIUS_USERS);
 #ifdef DBM
 		if (dbminit(buffer) != 0)
 #endif
@@ -876,9 +1121,9 @@ int user_find(char *name, VALUE_PAIR *request_pairs,
 			return 0;
 		}
 
-		if (dbm_find(name, request_pairs,
-			check_pairs, reply_pairs) < 0 ||
-			fallthrough(*reply_pairs)) {
+		r = dbm_find(name, request_pairs, check_pairs, reply_pairs);
+		if (r > 0) found = 1;
+		if (r <= 0 || fallthrough(*reply_pairs)) {
 
 			pairdelete(reply_pairs, PW_FALL_THROUGH);
 
@@ -887,6 +1132,7 @@ int user_find(char *name, VALUE_PAIR *request_pairs,
 			while ((r = dbm_find(buffer, request_pairs,
 			       check_pairs, reply_pairs)) >= 0 || i < 2) {
 				if (r > 0) {
+					found = 1;
 					if (!fallthrough(*reply_pairs))
 						break;
 					pairdelete(reply_pairs,PW_FALL_THROUGH);
@@ -912,6 +1158,9 @@ int user_find(char *name, VALUE_PAIR *request_pairs,
 			continue;
 
 		if (paircmp(request_pairs, pl->check) == 0) {
+			DEBUG2("  users: Matched %s at %d",
+			       pl->name, pl->lineno);
+			found = 1;
 			check_tmp = paircopy(pl->check);
 			reply_tmp = paircopy(pl->reply);
 			pairmove(reply_pairs, &reply_tmp);
@@ -930,7 +1179,7 @@ int user_find(char *name, VALUE_PAIR *request_pairs,
 	/*
 	 *	See if we succeeded.
 	 */
-	if (*reply_pairs == NULL)
+	if (!found)
 		return -1;
 
 	/*
@@ -945,6 +1194,9 @@ int user_find(char *name, VALUE_PAIR *request_pairs,
 				 *	numbers are stored in host order
 				 *	everywhere in this program.
 				 */
+#ifdef ASCEND_PORT_HACK
+				nas_port = ascend_port_number(nas_port);
+#endif
 				tmp->lvalue += nas_port;
 			}
 		}
@@ -963,7 +1215,7 @@ int user_find(char *name, VALUE_PAIR *request_pairs,
  *	Copy a data field from the buffer.  Advance the buffer
  *	past the data field.
  */
-static	void fieldcpy(char *string, char **uptr)
+static	void fieldcpy(char *string, int len, char **uptr)
 {
 	char	*ptr;
 
@@ -971,7 +1223,10 @@ static	void fieldcpy(char *string, char **uptr)
 	if(*ptr == '"') {
 		ptr++;
 		while(*ptr != '"' && *ptr != '\0' && *ptr != '\n') {
-			*string++ = *ptr++;
+			if (len-- > 0)
+				*string++ = *ptr++;
+			else
+				ptr++;
 		}
 		*string = '\0';
 		if(*ptr == '"') {
@@ -983,7 +1238,10 @@ static	void fieldcpy(char *string, char **uptr)
 
 	while(*ptr != ' ' && *ptr != '\t' && *ptr != '\0' && *ptr != '\n' &&
 						*ptr != '=' && *ptr != ',') {
-			*string++ = *ptr++;
+			if (len-- > 0)
+				*string++ = *ptr++;
+			else
+				ptr++;
 	}
 	*string = '\0';
 	*uptr = ptr;
@@ -998,18 +1256,19 @@ static	void fieldcpy(char *string, char **uptr)
 /*
  *	Parse the buffer to extract the attribute-value pairs.
  */
-static int userparse(char *buffer, VALUE_PAIR **first_pair)
+int userparse(char *buffer, VALUE_PAIR **first_pair)
 {
 	int		mode;
 	int		x;
-	char		attrstr[64];
-	char		valstr[64];
+	char		attrstr[256];
+	char		valstr[256];
 	char		*s;
-	DICT_ATTR	*attr;
+	DICT_ATTR	*attr = NULL;
 	DICT_VALUE	*dval;
 	VALUE_PAIR	*pair, *pair2;
 	struct tm	*tm;
 	time_t		timeval;
+	int		operator;
 
 	mode = PARSE_MODE_NAME;
 	while(*buffer != '\n' && *buffer != '\0') {
@@ -1023,42 +1282,78 @@ static int userparse(char *buffer, VALUE_PAIR **first_pair)
 
 		case PARSE_MODE_NAME:
 			/* Attribute Name */
-			fieldcpy(attrstr, &buffer);
+			fieldcpy(attrstr, sizeof(attrstr), &buffer);
 			if((attr = dict_attrfind(attrstr)) ==
 						(DICT_ATTR *)NULL) {
+#if 0 /* Be quiet. */
+				log(L_ERR|L_CONS, "unknown attribute %s",
+					attrstr);
+#endif
 				return(-1);
 			}
 			mode = PARSE_MODE_EQUAL;
 			break;
 
 		case PARSE_MODE_EQUAL:
-			/* Equal sign */
+			mode = PARSE_MODE_VALUE;
+			/* '=' Equal sign */
 			if(*buffer == '=') {
-				mode = PARSE_MODE_VALUE;
 				buffer++;
+				operator = PW_OPERATOR_EQUAL;
+			} else
+			/* '<' */
+			if(*buffer == '<') {
+				buffer++;
+				operator = PW_OPERATOR_LESS_THAN;
+			} else
+			/* '>' */
+			if(*buffer == '>') {
+				buffer++;
+				operator = PW_OPERATOR_GREATER_THAN;
+			} else
+			/* '!=' */
+			if(memcmp(buffer, "!=", 2) == 0) {
+				buffer += 2;
+				operator = PW_OPERATOR_NOT_EQUAL;
+			} else
+			/* '<=' */
+			if(memcmp(buffer, "<=", 2) == 0) {
+				buffer += 2;
+				operator = PW_OPERATOR_LESS_EQUAL;
+			} else
+			/* '>=' */
+			if(memcmp(buffer, ">=", 2) == 0) {
+				buffer += 2;
+				operator = PW_OPERATOR_GREATER_EQUAL;
 			}
 			else {
+				log(L_ERR|L_CONS, "expected `='");
 				return(-1);
 			}
 			break;
 
 		case PARSE_MODE_VALUE:
 			/* Value */
-			fieldcpy(valstr, &buffer);
-
+			fieldcpy(valstr, sizeof(valstr), &buffer);
 			if((pair = (VALUE_PAIR *)malloc(sizeof(VALUE_PAIR))) ==
 						(VALUE_PAIR *)NULL) {
 				log(L_CONS|L_ERR, "no memory");
 				exit(1);
 			}
+			/* No need for strncpy - same size */
 			strcpy(pair->name, attr->name);
 			pair->attribute = attr->value;
 			pair->type = attr->type;
+			pair->operator = operator;
 
 			switch(pair->type) {
 
 			case PW_TYPE_STRING:
-				strcpy(pair->strvalue, valstr);
+do_copy:
+				strncpy(pair->strvalue, valstr,
+					sizeof(pair->strvalue));
+				pair->strvalue[sizeof(pair->strvalue) - 1] = 0;
+				pair->length = strlen(pair->strvalue);
 				break;
 
 			case PW_TYPE_INTEGER:
@@ -1069,22 +1364,22 @@ static int userparse(char *buffer, VALUE_PAIR **first_pair)
 				if (attr->value == PW_NAS_PORT_ID) {
 					for(s = valstr; *s; s++)
 						if (!isdigit(*s)) break;
-					if (*s) {
-						pair->type = PW_TYPE_STRING;
-						strcpy(pair->strvalue, valstr);
-						break;
-					}
+					if (*s) goto do_copy;
 				}
 				if (isdigit(*valstr)) {
 					pair->lvalue = atoi(valstr);
+					pair->length = 4;
 				}
 				else if((dval = dict_valfind(valstr)) ==
 							(DICT_VALUE *)NULL) {
 					free(pair);
+					log(L_ERR|L_CONS, "unknown value %s",
+						valstr);
 					return(-1);
 				}
 				else {
 					pair->lvalue = dval->value;
+					pair->length = 4;
 				}
 				break;
 
@@ -1109,6 +1404,7 @@ static int userparse(char *buffer, VALUE_PAIR **first_pair)
 					}
 				}
 				pair->lvalue = get_ipaddr(valstr);
+				pair->length = 4;
 
 				/*
 				 *	Add an extra (hidden) attribute.
@@ -1122,6 +1418,7 @@ static int userparse(char *buffer, VALUE_PAIR **first_pair)
 				pair2->attribute = PW_ADD_PORT_TO_IP_ADDRESS;
 				pair2->type = PW_TYPE_INTEGER;
 				pair2->lvalue = x;
+				pair2->length = 4;
 				pairadd(first_pair, pair2);
 				break;
 
@@ -1134,10 +1431,15 @@ static int userparse(char *buffer, VALUE_PAIR **first_pair)
 #else /* TIMELOCAL */
 				pair->lvalue = (UINT4)mktime(tm);
 #endif /* TIMELOCAL */
+				pair->length = 4;
 				break;
 
 			default:
 				free(pair);
+#if 0 /* Yeah yeah */
+				log(L_ERR|L_CONS, "unknown attr. type %d",
+					pair->type);
+#endif
 				return(-1);
 			}
 			pairadd(first_pair, pair);
@@ -1156,7 +1458,7 @@ static int userparse(char *buffer, VALUE_PAIR **first_pair)
  *	Match a username with a wildcard expression.
  *	Is very limited for now.
  */
-static int matches(char *name, PAIR_LIST *pl, char *matchpart)
+static int matches(char *name, PAIR_LIST *pl, char *matchpart, int matchlen)
 {
 	int len, wlen;
 	int ret = 0;
@@ -1176,7 +1478,7 @@ static int matches(char *name, PAIR_LIST *pl, char *matchpart)
 
 		if (strncmp(pl->name, "DEFAULT", 7) == 0 ||
 		    strcmp(pl->name, name) == 0)
-			return !presufcmp(tmp, name, matchpart);
+			return !presufcmp(tmp, name, matchpart, matchlen);
 	}
 
 	/*
@@ -1185,13 +1487,15 @@ static int matches(char *name, PAIR_LIST *pl, char *matchpart)
 	if (strchr(pl->name, '*') == NULL &&
 	    (strncmp(pl->name, "DEFAULT", 7) == 0 ||
 	     strcmp(pl->name, name) == 0)) {
-		strcpy(matchpart, name);
+		strNcpy(matchpart, name, matchlen);
 		return 1;
 	}
 
 	/*
 	 *	Normally, we should return 0 here, but we
 	 *	support the old * stuff.
+	 *	FIXME: this doesn't support realsm yet, while
+	 *	presufcmp does!
 	 */
 	len = strlen(name);
 	wlen = strlen(wild);
@@ -1202,13 +1506,13 @@ static int matches(char *name, PAIR_LIST *pl, char *matchpart)
 		wild++;
 		wlen--;
 		if (wlen <= len && strcmp(name + (len - wlen), wild) == 0) {
-			strcpy(matchpart, name);
+			strNcpy(matchpart, name, matchlen);
 			matchpart[len - wlen] = 0;
 			ret = 1;
 		}
 	} else if (wild[wlen - 1] == '*') {
 		if (wlen <= len && strncmp(name, wild, wlen - 1) == 0) {
-			strcpy(matchpart, name + wlen - 1);
+			strNcpy(matchpart, name + wlen - 1, matchlen);
 			ret = 1;
 		}
 	}
@@ -1225,14 +1529,14 @@ static int matches(char *name, PAIR_LIST *pl, char *matchpart)
  */
 int hints_setup(VALUE_PAIR *request_pairs)
 {
-	char		newname[128];
+	char		newname[AUTH_STRING_LEN];
 	char		*name;
 	VALUE_PAIR	*add;
 	VALUE_PAIR	*last;
 	VALUE_PAIR	*tmp;
 	PAIR_LIST	*i;
 	int		do_strip;
-#ifdef NT_DOMAIN_HACK
+#if defined(NT_DOMAIN_HACK) || defined(SPECIALIX_JETSTREAM_HACK)
 	char		*ptr;
 #endif
 
@@ -1242,13 +1546,12 @@ int hints_setup(VALUE_PAIR *request_pairs)
 	/* 
 	 *	Check for valid input, zero length names not permitted 
 	 */
-	if ((tmp = pairfind(request_pairs, PW_USER_NAME)) == NULL) {
-		log(L_ERR,"no user name");
-		return -1;
-	}
-	name = tmp->strvalue;
+	if ((tmp = pairfind(request_pairs, PW_USER_NAME)) == NULL)
+		name = NULL;
+	else
+		name = tmp->strvalue;
 
-	if (name[0] == 0)
+	if (name == NULL || name[0] == 0)
 		/*
 		 *	Will be complained about later.
 		 */
@@ -1258,9 +1561,9 @@ int hints_setup(VALUE_PAIR *request_pairs)
 	/*
 	 *	Windows NT machines often authenticate themselves as
 	 *	NT_DOMAIN\username. Try to be smart about this.
+	 *	Note that sizeof(newname) == sizeof(name)
 	 *
-	 *	FIXME: should we handle this as a REALM ? (When we
-	 *	have implemented REALMS and proxy services, that is).
+	 *	FIXME: should we handle this as a REALM ?
 	 */
 	if ((ptr = strchr(name, '\\')) != NULL) {
 		strncpy(newname, ptr + 1, sizeof(newname));
@@ -1269,6 +1572,22 @@ int hints_setup(VALUE_PAIR *request_pairs)
 	}
 #endif /* NT_DOMAIN_HACK */
 
+#ifdef SPECIALIX_JETSTREAM_HACK
+	/*
+	 *	Specialix Jetstream 8500 24 port access server.
+	 *	If the user name is 10 characters or longer, a "/"
+	 *	and the excess characters after the 10th are
+	 *	appended to the user name.
+	 *
+	 *	Reported by Lucas Heise <root@laonet.net>
+	 */
+	if (strlen(name) > 10 && name[10] == '/') {
+		for (ptr = name + 11; *ptr; ptr++)
+			*(ptr - 1) = *ptr;
+		*(ptr - 1) = 0;
+	}
+#endif
+
 	/*
 	 *	Small check: if Framed-Protocol present but Service-Type
 	 *	is missing, add Service-Type = Framed-User.
@@ -1286,8 +1605,11 @@ int hints_setup(VALUE_PAIR *request_pairs)
 	}
 
 	for (i = hints; i; i = i->next) {
-		if (matches(name, i, newname))
+		if (matches(name, i, newname, sizeof(newname))) {
+			DEBUG2("  hints: Matched %s at %d",
+			       i->name, i->lineno);
 			break;
+		}
 	}
 
 	if (i == NULL) return 0;
@@ -1311,8 +1633,11 @@ int hints_setup(VALUE_PAIR *request_pairs)
 
 	if (do_strip) {
 		tmp = pairfind(request_pairs, PW_USER_NAME);
-		if (tmp)
+		if (tmp) {
+			/* Same size */
 			strcpy(tmp->strvalue, newname);
+			tmp->length = strlen(tmp->strvalue);
+		}
 	}
 
 	/*
@@ -1337,8 +1662,11 @@ static int huntgroup_match(VALUE_PAIR *request_pairs, char *huntgroup)
 	for (i = huntgroups; i; i = i->next) {
 		if (strcmp(i->name, huntgroup) != 0)
 			continue;
-		if (paircmp(request_pairs, i->check) == 0)
+		if (paircmp(request_pairs, i->check) == 0) {
+			DEBUG2("  huntgroups: Matched %s at %d",
+			       i->name, i->lineno);
 			break;
+		}
 	}
 
 	return (i != NULL);
@@ -1377,19 +1705,6 @@ int huntgroup_access(VALUE_PAIR *request_pairs)
 }
 
 
-/*
- *	Free a PAIR_LIST.
- */
-static void pairlist_free(PAIR_LIST *pl)
-{
-	while(pl) {
-		if (pl->name)  free(pl->name);
-		if (pl->check) pairfree(pl->check);
-		if (pl->reply) pairfree(pl->reply);
-		pl = pl->next;
-	}
-}
-
 /*
  *	Debug code.
  */
@@ -1446,6 +1761,7 @@ int read_clients_file(char *file)
 	CLIENT	*c;
 
 	clients_free(clients);
+	clients = NULL;
 
 	if ((fp = fopen(file, "r")) == NULL) {
 		log(L_CONS|L_ERR, "cannot open %s", file);
@@ -1455,11 +1771,14 @@ int read_clients_file(char *file)
 		lineno++;
 		if (buffer[0] == '#' || buffer[0] == '\n')
 			continue;
+
 		shortnm[0] = 0;
-		if (sscanf(buffer, "%s%s%s", hostnm, secret, shortnm) < 2) {
+		if (sscanf(buffer, "%127s%31s%31s", hostnm, secret, shortnm) < 2) {
 			log(L_ERR, "%s[%d]: syntax error", file, lineno);
 			continue;
 		}
+		if (shortnm[0] == '#') shortnm[0] = 0;
+
 		if ((c = malloc(sizeof(CLIENT))) == NULL) {
 			log(L_CONS|L_ERR, "%s[%d]: out of memory",
 				file, lineno);
@@ -1467,9 +1786,10 @@ int read_clients_file(char *file)
 		}
 
 		c->ipaddr = get_ipaddr(hostnm);
-		strcpy(c->secret, secret);
-		strcpy(c->shortname, shortnm);
-		strcpy(c->longname, ip_hostname(c->ipaddr));
+		strNcpy(c->secret, secret, sizeof(c->secret));
+		strNcpy(c->shortname, shortnm, sizeof(c->shortname));
+		strNcpy(c->longname, ip_hostname(c->ipaddr),
+			sizeof(c->longname));
 
 		c->next = clients;
 		clients = c;
@@ -1525,6 +1845,20 @@ static void nas_free(NAS *cl)
 	}
 }
 
+/*
+ *	Free a REALM list.
+ */
+static void realm_free(REALM *cl)
+{
+	REALM *next;
+
+	while(cl) {
+		next = cl->next;
+		free(cl);
+		cl = next;
+	}
+}
+
 /*
  *	Read the nas file.
  */
@@ -1539,6 +1873,7 @@ int read_naslist_file(char *file)
 	NAS	*c;
 
 	nas_free(naslist);
+	naslist = NULL;
 
 	if ((fp = fopen(file, "r")) == NULL) {
 		log(L_CONS|L_ERR, "cannot open %s", file);
@@ -1549,7 +1884,7 @@ int read_naslist_file(char *file)
 		if (buffer[0] == '#' || buffer[0] == '\n')
 			continue;
 		nastype[0] = 0;
-		if (sscanf(buffer, "%s%s%s", hostnm, shortnm, nastype) < 2) {
+		if (sscanf(buffer, "%127s%31s%31s", hostnm, shortnm, nastype) < 2) {
 			log(L_ERR, "%s[%d]: syntax error", file, lineno);
 			continue;
 		}
@@ -1560,9 +1895,10 @@ int read_naslist_file(char *file)
 		}
 
 		c->ipaddr = get_ipaddr(hostnm);
-		strcpy(c->nastype, nastype);
-		strcpy(c->shortname, shortnm);
-		strcpy(c->longname, ip_hostname(c->ipaddr));
+		strNcpy(c->nastype, nastype, sizeof(c->nastype));
+		strNcpy(c->shortname, shortnm, sizeof(c->shortname));
+		strNcpy(c->longname, ip_hostname(c->ipaddr),
+			sizeof(c->longname));
 
 		c->next = naslist;
 		naslist = c;
@@ -1572,6 +1908,90 @@ int read_naslist_file(char *file)
 	return 0;
 }
 
+/*
+ *	Read the realms file.
+ */
+int read_realms_file(char *file)
+{
+	FILE	*fp;
+	char	buffer[256];
+	char	realm[32];
+	char	hostnm[128];
+	char	opts1[32], opts2[32];
+	char	*p;
+	int	lineno = 0;
+	REALM	*c;
+
+	realm_free(realms);
+	realms = NULL;
+
+	if ((fp = fopen(file, "r")) == NULL) {
+#if 1 /* For now - realms file is not obligatory */
+		return 0;
+#else
+		log(L_CONS|L_ERR, "cannot open %s", file);
+		return -1;
+#endif
+	}
+	while(fgets(buffer, 256, fp) != NULL) {
+		lineno++;
+		if (buffer[0] == '#' || buffer[0] == '\n')
+			continue;
+		opts1[0] = opts2[0] = 0;
+		if (sscanf(buffer, "%31s%127s%31s%31s",
+		    realm, hostnm, opts1, opts2) < 2) {
+			log(L_ERR, "%s[%d]: syntax error", file, lineno);
+			continue;
+		}
+		if ((c = malloc(sizeof(REALM))) == NULL) {
+			log(L_CONS|L_ERR, "%s[%d]: out of memory",
+				file, lineno);
+			return -1;
+		}
+
+		if ((p = strchr(hostnm, ':')) != NULL) {
+			*p++ = 0;
+			c->auth_port = atoi(p);
+			c->acct_port = c->auth_port + 1;
+		} else {
+			c->auth_port = auth_port;
+			c->acct_port = acct_port;
+		}
+		if (strcmp(hostnm, "LOCAL") != 0)
+			c->ipaddr = get_ipaddr(hostnm);
+		strNcpy(c->realm, realm, sizeof(c->realm));
+		strNcpy(c->server, hostnm, sizeof(c->server));
+		/* Yes we could do this more intelligently */
+		c->striprealm = strcmp(opts1, "nostrip") != 0 ||
+				strcmp(opts2, "nostrip") != 0;
+		c->dohints    = strcmp(opts1, "hints") == 0 ||
+				strcmp(opts2, "hints") == 0;
+		c->next = realms;
+		realms = c;
+	}
+	fclose(fp);
+
+	return 0;
+}
+
+/*
+ *	Find a realm in the REALM list.
+ */
+REALM *realm_find(char *realm)
+{
+	REALM *cl;
+
+	for(cl = realms; cl; cl = cl->next)
+		if (strcmp(cl->realm, realm) == 0)
+			break;
+	if (cl) return cl;
+	for(cl = realms; cl; cl = cl->next)
+		if (strcmp(cl->realm, "DEFAULT") == 0)
+			break;
+	return cl;
+}
+
+
 
 /*
  *	Find a nas in the NAS list.
@@ -1628,7 +2048,27 @@ char *nas_name2(AUTH_REQ *authreq)
 	return ip_hostname(ipaddr);
 }
 
+/*
+ *	Return a short string showing the terminal server, port
+ *	and calling station ID.
+ */
+char *auth_name(AUTH_REQ *authreq, int do_cli)
+{
+	static char	buf[300];
+	VALUE_PAIR	*cli;
+	VALUE_PAIR	*pair;
+	int		port = 0;
 
+	if ((cli = pairfind(authreq->request, PW_CALLING_STATION_ID)) == NULL)
+		do_cli = 0;
+	if ((pair = pairfind(authreq->request, PW_NAS_PORT_ID)) != NULL)
+		port = pair->lvalue;
+
+	sprintf(buf, "from nas %.128s/S%d%s%.128s", nas_name2(authreq), port,
+		do_cli ? " cli " : "", do_cli ? cli->strvalue : "");
+
+	return buf;
+}
 
 #if USE_DBM
 /*
@@ -1639,8 +2079,7 @@ static int checkdbm(char *users, char *ext)
 	char buffer[256];
 	struct stat st;
 
-	strcpy(buffer, users);
-	strcat(buffer, ext);
+	sprintf(buffer, "%.200s%.50s", users, ext);
 
 	return stat(buffer, &st);
 }
@@ -1654,28 +2093,48 @@ int read_config_files()
 {
 	char buffer[256];
 
-	pairlist_free(users);
-	pairlist_free(huntgroups);
-	pairlist_free(hints);
+	pairlist_free(&users);
+	pairlist_free(&huntgroups);
+	pairlist_free(&hints);
 
-	sprintf(buffer, "%s/%s", radius_dir, RADIUS_USERS);
+	sprintf(buffer, "%.200s/%.50s", radius_dir, RADIUS_USERS);
 #if USE_DBM
-	if (!use_dbm &&
-		(checkdbm(users, ".dir") == 0 || checkdbm(users, ".db") == 0)
+	if (!use_dbm && users &&
+		(checkdbm(users->name, ".dir") == 0 ||
+		 checkdbm(users->name, ".db") == 0))
 			log(L_INFO|L_CONS, "DBM files found but no -b flag given - NOT using DBM");
 #endif
 	if (!use_dbm && (users = file_read(buffer, 1)) == NULL)
 		return -1;
-	sprintf(buffer, "%s/%s", radius_dir, RADIUS_HUNTGROUPS);
+
+	sprintf(buffer, "%.200s/%.50s", radius_dir, RADIUS_HUNTGROUPS);
 	huntgroups = file_read(buffer, 0);
-	sprintf(buffer, "%s/%s", radius_dir, RADIUS_HINTS);
+	sprintf(buffer, "%.200s/%.50s", radius_dir, RADIUS_HINTS);
 	hints      = file_read(buffer, 0);
-	sprintf(buffer, "%s/%s", radius_dir, RADIUS_CLIENTS);
+	sprintf(buffer, "%.200s/%.50s", radius_dir, RADIUS_CLIENTS);
 	if (read_clients_file(buffer) < 0)
 		return -1;
-	sprintf(buffer, "%s/%s", radius_dir, RADIUS_NASLIST);
+	sprintf(buffer, "%.200s/%.50s", radius_dir, RADIUS_NASLIST);
 	if (read_naslist_file(buffer) < 0)
 		return -1;
+	sprintf(buffer, "%.200s/%.50s", radius_dir, RADIUS_REALMS);
+	if (read_realms_file(buffer) < 0)
+		return -1;
+
+	if (cache_passwd) {
+		log(L_INFO, "HASH:  Reinitializing hash structures "
+			"and lists for caching...");
+		if(buildHashTable() < 0) {
+			log(L_ERR, "HASH:  unable to create user "
+				"hash table.  disable caching and run debugs");
+			return -1;
+		}
+		if (buildGrpList() < 0) {
+			log(L_ERR, "HASH:  unable to cache groups file.  "
+				"disable caching and run debugs");
+			return -1;
+		}
+	}
 
 	return 0;
 }
