@@ -37,7 +37,7 @@ char version_sccsid[] =
  *  would like a range of versions allocated for your use.
  */
 
-#define		VERSION		"cistron-1.5.4 21-Oct-1997"
+#define		VERSION		"cistron-1.6.1 19-Oct-1999"
 
 /*
  *	Display the revision number for this program
@@ -58,9 +58,24 @@ void version(void)
 #if defined(NOSHADOW)
 	fprintf(stderr," NOSHADOW");
 #endif
+#if defined(OSFC2)
+	fprintf(stderr," OSFC2");
+#endif
 #if defined(NT_DOMAIN_HACK)
 	fprintf(stderr," NT_DOMAIN_HACK");
 #endif
+#if defined(SPECIALIX_JETSTREAM_HACK)
+	fprintf(stderr," SPECIALIX_JETSTREAM_HACK");
+#endif
+#if defined(NOCASE)
+	fprintf(stderr," NOCASE");
+#endif
+#if defined(ATTRIB_NMC)
+	fprintf(stderr, " ATTRIB_NMC");
+#endif
+#if defined(COMPAT_1543)
+	fprintf(stderr, " COMPAT_1543");
+#endif
 
 	/* here are all the system definitions compilation uses */
 #if defined(__alpha)
